# Takes an SCI0 resource file(s), strips the text resources from it, optionally calls
# a spell checker & grammar checker and then appends the text resources back into 
# the resource file & recreates the resource.map file

# Preparation:  Copy your resource.xxx files into the same directory as this script 
#               (in case something goes wrong - you won't nuke your only copies of the resource files)
#               Developed with Strawberry Perl 5.12.0.  The only additional perl dependency is the 
#               Win32::Console::ANSI libaray (which I obtained via the CPAN client)

# Usage:
# 1.  Dump the resource.xxx file text resources out (which strips the resource.xxx file of all text resources into a new file)
# 2.  Run the spell check/grammar check resources against the resulting text resource dump
# 3.  Append the modified text resource file back into the stripped resource.xxx file
# 4.  Repeat steps 1-3 for each resource file for the game
# 5.  Create a new resource.map file based on the newly completed resource.xxx file

# Spell check external dependencies:
# 1.  Install the GNU ASpell library (get the 'Full Installer' for windows & the 'English' precompiled dictionary 2 separate downloads)
# Note: This application will first look for the aspell app to be installed in the local directory, then in the "Program Files" directory,
#     then the entire drive.

# Grammar check external dependencies:
# 1.  Download LanguageTool java application
# 2.  Rename the ".oxt" archive to ".zip" and unzip it - I recommend putting it in a subfolder where this application is located
# 3.  If you're using Java 5.0 unzip the standalone-libs.zip that will be created (into the same directory)
# Note: This application will first look for the jar to be installed in the local directory, then in the "Program Files" directory,
#     then the entire drive.

# Known bugs:
#   - If dumped text resource file exceeds 64K, line numbers get fouled up in LanguageTool (version 1.1) - heed it no mind         

# To Do:
#   -  Add support for compressed text resources
#   -  Rework interactive menus to be more intuitive

# Do whatever you want with this file, except remove my name.
# Jeremiah Nellis, nellisjp@hotmail.com
# Version 1.0 - Oct 11th 2010

use strict;
use warnings;

use Term::ANSIColor qw(:constants);
use Term::UI;
use Term::ReadLine;
use File::Find;
use Cwd;

use Win32::Console::ANSI;

no warnings 'File::Find';

my $hex;
my $data;
my $hexVal1;
my $hexVal2;
my $hexVal;
my $file;
my $rest;
my $i = 0;

my $junk;

my $restText;
my $resourceNumber = "";

my $strippedFile;

my @mapFile;

sub ascii_to_hex($);
sub hex_to_ascii($);
sub dumpResources($$$);
sub spellCheck($);
sub createGrammarCheckFile($$);
sub addMap($$);
sub writeMap($);
sub appendResourceFile($$);
sub parseGrammarCheckFile($$);

my $term = Term::ReadLine->new();
while(1) {

	my $menu = $term->get_reply(
	   prompt => 'What do you want to do? ',
	   choices => ["Dump text resources (from a resource.001 file)", 
				   "Spell check dump file", 
				   "Grammar check dump file", 
				   "Append modified text resources to stripped resource file from dump",
                   "Create new map file from resource file(s)",				   
				   "Exit"],
	);

	if($menu eq "Dump text resources (from a resource.001 file)") {
	   my $inputResFileName = $term->get_reply(
          prompt => 'Input resource file',
		  default => 'resource.001'
       );	   
	   my $outputTextFileName = $term->get_reply(
	      prompt => 'Output text file',
		  default => 'textResources.txt'
	   );
	   my $outputResFileName = $term->get_reply(
	      prompt => 'Output resource file (stripped of text resources):',
		  default => 'resource.001.new'
	   );
	   dumpResources($inputResFileName, $outputTextFileName, $outputResFileName);	   
	}
	
	if($menu eq "Spell check dump file") {
	   my $inputTextName = $term->get_reply(
	      prompt => 'Input file name:',
		  default => 'textResources.txt'	   
	   );
	   spellCheck($inputTextName);
	}
	
	if($menu eq "Grammar check dump file") {
	   my $inputTextFile = $term->get_reply(
	      prompt => 'Input text file name:',
		  default => 'textResources.txt'			  
	   );	   		   
	   createGrammarCheckFile($inputTextFile, "grammarOutput.txt");
	   parseGrammarCheckFile("grammarOutput.txt", $inputTextFile);
	}
	
	if($menu eq "Append modified text resources to stripped resource file from dump") {
	   my $inputFileName = $term->get_reply(
	      prompt => 'Input text file name:',
		  default => 'textResources.txt'			  
	   );
	   my $appendResFileName = $term->get_reply(
	      prompt => 'Append resource file name:',
		  default => 'resource.001.new'			  
	   );		   	   
	   appendResourceFile($inputFileName, $appendResFileName);
	}
	
	if($menu eq "Create new map file from resource file(s)") {
	   my $submenu = "";
	
	   while($submenu ne "Write out map") {
	      $submenu = $term->get_reply(
	         prompt => 'What do you want to do? ',
	         choices => ["Add resource file to map", 
		  		         "Write out map",
						 "Return to main menu"],
	      );
		  
	      if ($submenu eq "Add resource file to map") {
		  	 my $inputResFileName = $term->get_reply(
                prompt => 'Input resource file',
		        default => 'resource.001.new'
             );	
			 my $resourceNumber = $term->get_reply(
			    prompt => 'Enter resource number for this input file: '
			 );
	         addMap($inputResFileName, $resourceNumber);
	      } 
		  
		  if ($submenu eq "Write out map") {
		     my $outputMapFileName = $term->get_reply(
	            prompt => 'Output resource.map file name:',
		        default => 'resource.map.new'			  
	         );
	         writeMap($outputMapFileName);
		  }
		  
		  if($submenu eq "Return to main menu") {
		     # this is ugly...
		     $submenu = "Write out map";
		  }
	    }
	
	}
	
	if($menu eq "Exit") {
	  print "...bye\n";
	  exit(0);
	} else {
	  print "...returning to main menu:\n\n";
	}	
};

exit(0);
# End of main

sub ascii_to_hex ($) {
	## Convert each ASCII character to a two-digit hex number.
	(my $str = shift) =~ s/(.|\n)/sprintf("%02lx", ord $1)/eg;
	return $str;
}

sub hex_to_ascii ($) {
	## Convert each two-digit hex number back to an ASCII character.
	(my $str = shift) =~ s/([a-fA-F0-9]{2})/chr(hex $1)/eg;
	return $str;
}



sub dumpResources($$$) {
    my($inputResFile, $outputTextFile, $outputResFile) = @_;
	
	# Read whole file in
	$data = do {
		local $/ = undef;
		open my $fh, $inputResFile
			or die "could not open $file: $!";
		binmode $fh;
		<$fh>;
	};


	print "Dumping resources...\n";

	# Open file to dump text resources to
	open(OUTFILE,">" . $outputTextFile) || die;


	$rest = $data;
	$strippedFile = $data;

	while ($rest) {
	   # get the 'header' for the resource (8 bytes)
	   #   pattern:  (00 - leading char) | res # - hex + hex 18 through 1B (*256 offset) | comp. size (2) | decomp. size (2) | 00 00 
	   $hexVal = substr($rest, 0, 8);

	   my $resLen = int(ord(substr($hexVal,2,1))) + int(ord(substr($hexVal,3,1))) * 256 + 4;

	   # strip out only this text resource from the string   
	   $restText = substr($rest,8,$resLen - 8);
	   
	   $rest = substr($rest,$resLen);
	  
	   my $stripText = quotemeta($hexVal . $restText);
	   $strippedFile =~ s/($stripText)//;

	   my $resourceType = unpack("B32", pack("H*",substr($hexVal,1,1) . substr($hexVal,0,1)) ); 

       my $hex1 = unpack("B16", substr($hexVal,0,1));
	   my $hex2 = unpack("B16", substr($hexVal,1,1));
	   
	   $resourceType = substr($hex2, 0,5);
	   $resourceNumber = $hex1 . " " . substr($hex2,5);
	   
	   $resourceType = unpack("N", pack("B32", substr("0" x 32 . $resourceType, -32)));
	   my $offset = int(unpack("N", pack("B32", substr("0" x 32 . substr($hex2,5), -32)))) * 256;
	   $resourceNumber = int(unpack("N", pack("B32",substr("0" x 32 . $hex1, -32)))) + $offset;
	   my $compType =  substr(unpack("H2", substr($hexVal,6,1)), 1,1);
	   
	  # Compression types:  0 - none, 1 - LZW, 2 - COMP-3


 	  # print " Resource id: " . ascii_to_hex(substr($hexVal,0,2)) . " Res length = " . $resLen
	  # . " resType = " . $resourceType . " resNum = " . $resourceNumber 
	  # . " compType = " . $compType . "\n";

	   	  
	  # Currently, only support text resources with NO compression
	   if($compType != 0 || $resourceType != 3) {
	      next;
	   }
	   
	   print OUTFILE "/********** Text Resource #" . $resourceNumber 
				   . ", Resource Id: " . ascii_to_hex(substr($hexVal,0,2)) . " *********/\n";
	   
	   while ($restText) {
			 (my $text, $restText) = split(/\x00/, $restText, 2);
				 #print "Res#: " . ord(substr($hexVal,0,1)) . " RType: " . ord(substr($hexVal,1,1)) . " #: " . $i . "  ";
				 
				 # Find & replace all carriage returns with "\n"
				 $text =~ s/\x0A/\\n/;
			 print OUTFILE $text . "\n";
	   }   

	}

	print OUTFILE "/********** End of file **********/";
	close(OUTFILE);

	open(NEWFILE,">" . $outputResFile) || die;
	binmode NEWFILE;
	print NEWFILE $strippedFile;
	close(NEWFILE);
}


sub spellCheck($) {
	my($inputFile) = @_;

	print "Spell checking...\n";
	my $cwd = getcwd();
	my $aspell;
	
	print "Locating aspell executable.  Looking in current directory first.\n";
	
	my $dir = $cwd;
	find(\&thisDirSpell, $dir);
	sub thisDirSpell {
	   no warnings;
	   if ($_ eq "aspell.exe") {
	      print "Using aspell.exe found at: " . $File::Find::name . "\n";
		  $aspell = $File::Find::name;
		  goto endFindSpell;
	   }
	};
	
    print "Looking in " . $ENV{SYSTEMDRIVE} . "\\Program Files" . " directory next..\n";
	
	$dir = $ENV{SYSTEMDRIVE} . "\\Program Files";
	find(\&knownDirSpell, $dir);
	sub knownDirSpell {
	   no warnings;
	   if ($_ eq "aspell.exe") {
	      print "Using aspell.exe found at: " . $File::Find::name . "\n";
		  $aspell = $File::Find::name;
		  goto endFindSpell;
	   }
	};
	
	$dir = $ENV{SYSTEMDRIVE} . "\\";
    print "Didn't find it.  Now checking the entire system drive (this may take a while)...\n";
	
	find(\&entireDriveSpell, $dir);
	sub entireDriveSpell {
	   no warnings;
	   if ($_ eq "aspell.exe") {
	      print "Using aspell.exe found at: " . $File::Find::name . "\n";
		  $aspell = $File::Find::name;
		  goto endFindSpell;
	   }
	};
	
	print "Could not find aspell.exe...";
	return;
	
	endFindSpell:
	chdir $cwd;
	
	my @args = ($aspell, "-c", $inputFile);
	system(@args) == 0 or die "system @args failed: $?";
}

sub createGrammarCheckFile($$) {
	my($inputFile, $outputFile) = @_;
	my @args;
	my $languageToolJar;

	my $cwd = getcwd();
	
	print "Locating LanguageTool.jar.  Looking in current directory first.\n";
	
	my $dir = $cwd;
	find(\&thisDir, $dir);
	sub thisDir {
	   no warnings;
	   if ($_ eq "LanguageTool.jar") {
	      print "Using LanguageTool found at: " . $File::Find::name . "\n";
		  $languageToolJar = $File::Find::name;
		  goto endFind;
	   }
	};
	
	$dir = substr($cwd,0,3);
    print "Didn't find it.  Now checking the entire drive (this may take a while)...\n";
	
	find(\&entireDrive, $dir);
	sub entireDrive {
	   no warnings;
	   if ($_ eq "LanguageTool.jar") {
	      print "Using LanguageTool found at: " . $File::Find::name . "\n";
		  $languageToolJar = $File::Find::name;
		  goto endFind;
	   }
	};
	
    print "Could not find LanguageTool.jar...";
	return;
	
	endFind:
	chdir $cwd;
	
	my $checkWhitespace = $term->ask_yn(
          prompt => 'Enable whitespace checking?',
		  default => 'y'	
	);
	
	print "Grammar checking...\n";

	if ($checkWhitespace) {
	     @args = ("cmd", "/c", "java.exe", "-jar", $languageToolJar,
	              "-d", "EN_QUOTES", $inputFile, ">", $outputFile);
    } else {
	     @args = ("cmd", "/c", "java.exe", "-jar", $languageToolJar,
	              "-d", "EN_QUOTES,WHITESPACE_RULE",  $inputFile, ">", $outputFile);	
	}
	system(@args) == 0 or die "system @args failed: $?";
}

sub addMap($$) {
    my($inputResFile, $resourceNumber) = @_;
	print "Add resource indexes to map array...\n";

	my $data = do {
	    local $/ = undef;
	    open my $fh, $inputResFile
		or die "could not open $file: $!";
	    binmode $fh;
	    <$fh>;
	};

	$rest = $data;
	my $filePos = int(0);

	while ($rest) {

	   my $hexVal = substr($rest,0,8);
	   my $resLength = int(ord(substr($hexVal,2,1))) + int(ord(substr($hexVal,3,1))) * 256 - 5;
	   my $mapValue;
	   
	   # Get the header & the body for this header
	   my $hexBody = substr($rest, 8, $resLength);
	   
	   if(length($rest) >= $resLength + 9) {
	      $rest = substr($rest, 9 + $resLength);
	   } else {
	      $rest = undef;
	   }

	   my $hexBodyLen = sprintf("%x", $filePos);

	   # left pad uneven hex values
	   if (length($hexBodyLen)%2==1) {
		   $hexBodyLen = "0" . $hexBodyLen;
	   }

	   # reverse the hex values and right pad with extra 00 if necessary
	   if(length($hexBodyLen) == 6) {
		   $hexBodyLen = substr($hexBodyLen,4,2) . substr($hexBodyLen,2,2) . substr($hexBodyLen,0,2);
	   }

	   if(length($hexBodyLen) == 4) {
		   $hexBodyLen = substr($hexBodyLen,2,2) . substr($hexBodyLen,0,2);
		   $hexBodyLen = $hexBodyLen . "00";
	   } 

	   if(length($hexBodyLen) == 2) {
		   $hexBodyLen = $hexBodyLen . "0000";
	   }

	   # write out the resource number/type
	   $mapValue = substr($hexVal,0,2);

	   # write out the length
	   $mapValue = $mapValue . pack('H*',$hexBodyLen);


	   # conver the resource number to binary, then to hex
	   my $resNum = unpack("B*", pack("N", $resourceNumber));
	   $resNum = substr($resNum,length($resNum) - 6) . "00";
       my $hex =  sprintf('%X', oct("0b$resNum"));
      	   
	   if(length($hex) == 1) {
	      $hex = "0" . $hex;
	   }
       
	   $mapValue = $mapValue . pack('H*', $hex);

	   push(@mapFile, $mapValue);
	  
	   # calculate the absolute position for the next resource
	   $filePos = $filePos + length($hexBody) + 9;

	   $i++;
    }
}

sub writeMap($) {
    my($outputMapFile) = @_;
	my $mapEntry;
	
	print "Creating new map file...\n";

	open(RESOURCEMAP, ">" . $outputMapFile) || die;
	binmode RESOURCEMAP;

    foreach $mapEntry (@mapFile) {    
       print RESOURCEMAP $mapEntry;
	}   

	@mapFile = ();
	
	# Write out the file footer
	print RESOURCEMAP pack('H*', 'FFFFFFFFFFFF');
	
	close(RESOURCEMAP);
}

sub appendResourceFile ($$) {
    my($inputTextFile, $outputResFile) = @_;
    my $line;
    my $resourceId;
    my $junk;
    my @resource;
    my $resourceLength = 0;
    my $resLine;
	my $firstResource = 1;
	

   # windows carriage return only!
    local $/ = "\r\n"; 

    print "Appending text resources to resource file...\n";    

    open TEXTFILE, $inputTextFile or die "Can't open $inputTextFile $!\n";
    binmode TEXTFILE;
    open RESOURCEFILE, ">>" . $outputResFile || "Can't out $outputResFile $!\n"; 
    binmode RESOURCEFILE;
 
    foreach $line (<TEXTFILE>) {
       chomp($line);              # remove the newline from $line.

       # new resource found (/*** found)
       if ($line =~ m/\/\*\*\*[\s\S]*/) {
	   
		   if (@resource) {
		   		   
			   # Find the total size of the array
			   foreach $resLine (@resource) {    
                   # Find & replace all "\n" to carriage returns
                   $resLine =~ s/\\n/\x0A/;        
			       $resourceLength = $resourceLength + length($resLine) + 1;
			   }			   

               # Write out the resource id/type, compressed size & decompressed size
			   if ($firstResource == 0) {
                   print RESOURCEFILE pack('H*', '00');
               } else {
                   $firstResource = 0;
               }	           
			   
			   print RESOURCEFILE pack('H*', $resourceId);
			   
			   # compressed size
               my $compressedLength = int($resourceLength) + int(4); # add length of header + '00' footer
               my $compressedSize1 = $compressedLength % 256;
               my $compressedSize2 = int($compressedLength / 256);
               print RESOURCEFILE pack('C', $compressedSize1);
			   print RESOURCEFILE pack('C', $compressedSize2);

			   # decompressed size
               my $decompressedLength = int($resourceLength);
               my $decompressedSize1 = $decompressedLength % 256;
               my $decompressedSize2 = int($decompressedLength / 256);
               print RESOURCEFILE pack('C', $decompressedSize1);
			   print RESOURCEFILE pack('C', $decompressedSize2);

			   # compression type
			   print RESOURCEFILE pack('H*', '00');
           
			   # Write out each text string in this resource
			   foreach $resLine (@resource) { 
                               # Find & replace all "\n" to carriage returns
                               $resLine =~ s/\\n/\x0A/;

			       print RESOURCEFILE pack('H*', '00');
			       print RESOURCEFILE $resLine;
			   }		   
		   }

		   # Clear out the array for the next resource
		   @resource=();
		   $resourceLength = 0;

           ($junk, $resourceId) = split /Resource Id\:[\s]/, $line;
         
           unless ($resourceId) {
               next;
           }

           ($resourceId, $junk) = split /[\s]/, $resourceId;
           #print "Resource id = " . $resourceId . "\n";
       } else {
	       push(@resource, $line);
       }
    } 
	
	# Write a '00' for the last resource
    print RESOURCEFILE pack('H*', '00');
	
    close TEXTFILE;
    close RESOURCEFILE;
}

sub parseGrammarCheckFile($$) {
    my($grammarFile, $textResourceFile) = @_;
	my $grammarCheck;
	my $location;
	my $message;
	my $suggestion;
	my $text;
	my $line;
	my $junk;
	my $junk2;
	my $column;
	my $lineNo;
	my $replacement;
	my @fileReplacements;
	my $i = 0;

	print "Parsing grammar check results...\n";
	
	my $rest = do {
	    local $/ = undef;
	    open my $fh, $grammarFile
		or die "could not open $file: $!";
	    binmode $fh;
	    <$fh>;
	};
	
	my $textResource = do {
	    local $/ = undef;
	    open my $fh, $textResourceFile
		or die "could not open $file: $!";
	    binmode $fh;
	    <$fh>;
	};
	

	while ($rest) {
	   ($grammarCheck, $rest) = split /[\d]+\.\)[\s]/, $rest, 2;

	   # strip out the last line
	   $grammarCheck =~ s/Time\:[\s][\d]+[\s\S]+//;
	
	   # split output into multiple parts...
	   if ($grammarCheck =~ m/Suggestion\:/) {
	     ($location, $message, $suggestion, $text) = split /[\x0D][\x0A]/, $grammarCheck;
	   } else {
	   	 ($location, $message, $text) = split /[\x0D][\x0A]/, $grammarCheck;
		 $suggestion = undef;
	   }
	   
	   unless ($text) {
	      next;
	   }
	   
	   # Get the adjusted column number
	   ($junk, $column, my $suspectLength, $junk2) = split /\.[\s\.][\s\.]([\s]+)(\^+)/, $text, 3;

	   # Remove trailing whitespace & carats
	   ($text, $junk) = split /[\s]+\^+/, $text;
	   
	   $column = length($column) - 1;	
       $suspectLength = length($suspectLength);	   
 
	   my $beginningText = substr($text,0,$column);
	   my $suspectText = substr($text,$column,$suspectLength);
	   my $endingText = substr($text, $column + $suspectLength);

	   print $location . "\n";
	   print $message . "\n";
	   if (defined $suggestion) {
	      print $suggestion . "\n";
	   }
	  # print "Text = " . $text . "\n";

	   print $beginningText;
	   print WHITE, ON_BLUE, $suspectText;
	   print RESET;
	   print $endingText;
	   print "\n\n";
	   
	   print "Enter replacement text (enter for no replacement).\n";

	   $replacement = <>;

           $replacement = substr($replacement,0,length($replacement) - 1);
 
           if (length($replacement) == 0) {   
              print "No replacement being made, skipping this...\n\n";
              next;
           }

	   print "\n\n";

		   # Strip off the beginning & ending 3 characters (usually ellipses)
		   if (length($beginningText) > 3) {
		     $beginningText = substr($beginningText, 3);
		   }
		   
		   if (length($endingText) > 3) {
		     $endingText = substr($endingText, 0, length($endingText) - 3);
		   }

           $beginningText = quotemeta($beginningText);
              $endingText = quotemeta($endingText);


           $beginningText =~ s/\\\x20/\[\^\\S\]\+/g;
              $endingText =~ s/\\\x20/\[\^\\S\]\+/g;

           $suspectText = quotemeta($suspectText);
		   
		   # Fix whitespace
		   $suspectText =~ s/\\ /\[\^\\S\]\+/g;
		   
           # Add pair of parens so we can get the location of the matched text later
           my $regex = $beginningText . "(" . $suspectText . ")" . $endingText;  

           if ($textResource =~ m/$regex/) {
		      my $matchLoc = $-[1];
              #print "Match found at: " . $matchLoc . "\n";
              #print substr($textResource, $matchLoc, $suspectLength) . "\n";
              #$textResource = substr($textResource,0,$matchLoc) 
              #              . $replacement
              #              . substr($textResource,$matchLoc + $suspectLength);            
			  $fileReplacements[$i]{'replacementText'} = $replacement;
			  $fileReplacements[$i]{'originalText'} = substr($textResource, $matchLoc, $suspectLength);
			  $fileReplacements[$i]{'startPos'} = $matchLoc;
              $i++;
           } else {
              print "regex string = " . $regex . "\n"; 
              print "Uh-oh, couldn't find a match!\n\n";
           }

           print "\n\n";          
	}

	my $offset = 0;
    for ($i=0; $i < scalar(@fileReplacements); $i++) {
	   print " Replace text = " . $fileReplacements[$i]{'replacementText'};
	   print " Original text = " . $fileReplacements[$i]{'originalText'};
	   print " Starting position = " . $fileReplacements[$i]{'startPos'};
	   print "\n";

	   $textResource = substr($textResource, 0, $offset + $fileReplacements[$i]{'startPos'})
	                 . $fileReplacements[$i]{'replacementText'}
					 . substr($textResource, $offset + $fileReplacements[$i]{'startPos'} + length($fileReplacements[$i]{'originalText'}));
	   $offset = $offset + length($fileReplacements[$i]{'replacementText'}) - length($fileReplacements[$i]{'originalText'});
	}
	
    # Open file to dump text resources to
    open(OUTFILE,">" . $textResourceFile) || die;
	binmode(OUTFILE);
    print OUTFILE $textResource;
    close OUTFILE; 
}
